/* $Id: ifactor1.c,v 2.0.0.9 1998/06/09 11:47:19 belabas Exp belabas $ */
#include "pari.h"

/*********************************************************************/
/**                                                                 **/
/**                        PSEUDO PRIMALITY                         **/
/**                                                                 **/
/*********************************************************************/
static GEN sqrt1, sqrt2, t1, t;
static long r1;

static GEN
init_miller(GEN n)
{
  t=subis(n,1); r1=vali(t); t1 = shifti(t,-r1);
  sqrt1=cgeti(lg(t)); sqrt1[1]=evalsigne(0)|evallgefint(2);
  sqrt2=cgeti(lg(t)); sqrt2[1]=evalsigne(0)|evallgefint(2);
  return (signe(n) < 0)? absi(n): n;
}

/* is n strong pseudo-prime for base a ? `End matching' (check for square
 * roots of -1) added by GN */
/* TODO: If ends do mismatch, then we have factored n, and this information
   should somehow be made available to the factoring machinery. --GN */
static int
bad_for_base(GEN n, GEN a)
{
  GEN c2, c = powmodulo(a,t1,n);
  long r;

  if (!is_pm1(c) && !egalii(t,c)) /* go fishing for -1, not for 1 */
  {
    for (r=r1-1; r; r--)	/* (this saves one squaring/reduction) */
    {
      c2=c; c=modii(sqri(c),n);
      if (egalii(t,c)) break;
    }
    if (!r) return 1;
    /* sqrt(-1) seen, compare or remember */
    if (signe(sqrt1))		/* we saw one earlier: compare */
    {
      /* check if too many sqrt(-1)s mod n */
      if (!egalii(c2,sqrt1) && !egalii(c2,sqrt2)) return 1; 
    }
    else { affii(c2,sqrt1); affii(subii(n,c2),sqrt2); } /* remember */
  }
  return 0;
}

/* Miller-Rabin test for k random bases */
long
millerrabin(GEN n, long k)
{
  long r,i,av2, av = avma;
  
  if (!signe(n)) return 0;
  /* If |n| <= 3, check if n = +- 1 */
  if (lgefint(n)==3 && (ulong)(n[2])<=3) return (n[2] != 1); 

  if (!mod2(n)) return 0;
  n = init_miller(n); av2=avma;
  for (i=1; i<=k; i++)
  {
    do r = smodsi(mymyrand(),n); while (!r);
    if (bad_for_base(n, stoi(r))) { avma=av; return 0; }
    avma=av2;
  }
  avma=av; return 1;
}

/* As above for k bases taken in pr (i.e not random).
 * We must have k<=11 and |n|>2 (not checked), or k in {16,17} to select
 * some special sets of bases.
 * 
 * By computations of Gerhard Jaeschke, `On strong pseudoprimes to several
 * bases', Math.Comp. 61 (1993), 915--926  (see also Chris Caldwell's Prime
 * Number Pages at http://www.utm.edu/research/primes/prove2.html),  we have:
 * 
 * k == 4  (bases 2,3,5,7)  correctly detects all composites
 *    n <     118 670 087 467 == 172243 * 688969  with the single exception of
 *    n ==      3 215 031 751 == 151 * 751 * 28351,
 * 
 * k == 5  (bases 2,3,5,7,11)  correctly detects all composites
 *    n <   2 152 302 898 747 == 6763 * 10627 * 29947,
 * 
 * k == 6  (bases 2,3,...,13)  correctly detects all composites
 *    n <   3 474 749 660 383 == 1303 * 16927 * 157543,
 * 
 * k == 7  (bases 2,3,...,17)  correctly detects all composites
 *    n < 341 550 071 728 321 == 10670053 * 32010157,
 * and even this limiting value is caught by an end mismatch between bases
 * 2 and 5 (or 5 and 17).
 * 
 * Moreover, the four bases chosen at
 * 
 * k == 16  (2,13,23,1662803)  will correctly detect all composites up
 * to at least 10^12, and the combination at
 * 
 * k == 17  (31,73)  is foolproof for odd composites e.g. in the ranges
 *    2^32 < n < 
 * with only  exceptions, all of which have a prime factor < 100. --GN
 * (DATA TO BE COMPLETED)
 */
int				/* no longer static -- needed in mpqs.c */
miller(GEN n, long k)
{
  long r,i,av2, av = avma;
  static long pr[] =
    {0, 2,3,5,7,11,13,17,19,23,29, 31,73, 2,13,23,1662803};
  long *p;

  if (!mod2(n)) return 0;
  if (k==16) { p=pr+13; k=4; }
  else if (k==17) { p=pr+11; k=2; }
  else { p=pr; }
  n = init_miller(n); av2=avma;
  for (i=1; i<=k; i++)
  {
    r = smodsi(p[i],n); if (!r) break;
    if (bad_for_base(n, stoi(r))) { avma = av; return 0; }
    avma=av2;
  }
  avma=av; return 1;
}

GEN
nextprime(GEN n)
{
  long toadd,av1,av2, av = avma;

  if (typ(n) != t_INT) n=gceil(n); /* accept arguments in R --GN */
  if (typ(n) != t_INT) err(arither1);
  if (signe(n) <= 0) { avma=av; return gdeux; }
  if (lgefint(n) <=3)
  { /* check if n<=3 */
    ulong k;
    if ((k=n[2])<=2) { avma=av; return gdeux; }
    if (k == 3) return stoi(3);
  }
  /* here n>3 */
  switch(smodis(n,6))
  {
    case 0: n=addsi(1,n); toadd=4; break;
    case 1: toadd=4; break;
    case 2: n=addsi(3,n); toadd=2; break;
    case 3: n=addsi(2,n); toadd=2; break;
    case 4: n=addsi(1,n); toadd=2; break;
    case 5: toadd=2; break;
    default: return NULL; /* not reached */
  }
  av2 = av1 = avma;
  for(;;)
  {
    if (miller(n,10)) break;
    av1=avma; n=addsi(toadd,n); toadd = 6-toadd;
  }
  if (av1!=av2) return gerepile(av,av1,n);
  return (av1==av)? icopy(n): n;
}

GEN
precprime(GEN n)
{
  long tosub,av1,av2, av = avma;

  if (typ(n) != t_INT) n=gfloor(n); /* accept arguments in R --GN */
  if (typ(n) != t_INT) err(arither1);
  if (signe(n)<=0) { avma=av; return gzero; }
  if (lgefint(n) <= 3)
  { /* check if n<=4 */
    ulong k;
    if ((k=n[2])<=1) { avma=av; return gzero; }
    if (k <= 3) { avma=av; return stoi(k); }
    if (k == 4) { avma=av; return stoi(3); }
  }
  /* here n>4 */
  switch(smodis(n,6))
  {
    case 0: n=addsi(-1,n); tosub=4; break;
    case 1: tosub=2; break;
    case 2: n=addsi(-1,n); tosub=2; break;
    case 3: n=addsi(-2,n); tosub=2; break;
    case 4: n=addsi(-3,n); tosub=2; break;
    case 5: tosub=4; break;
    default: return NULL; /* not reached */
  }
  av2 = av1 = avma;
  for(;;)
  {
    if (miller(n,10)) break;
    av1=avma; n=addsi(-tosub,n); tosub = 6-tosub;
  }
  if (av1!=av2) return gerepile(av,av1,n);
  return (av1==av)? icopy(n): n;
}

/***********************************************************************/
/**                                                                   **/
/**                        FACTORIZATION (ECM)                        **/
/**   Integer factorization using the elliptic curves method (ECM).   **/
/**   ellfacteur() returns a non trivial factor of N, assuming N>0,   **/
/**   is composite, and has no small prime divisor (P^-(N) > 65.536). **/
/**                                                                   **/
/***********************************************************************/
static GEN N, gl, *W;
#define nbc 10

static int
sisprime(long p)
{
  long av = avma, i = miller(stoi(p),5);
  avma = av; return i;
#if 0
  byteptr d = diffptr;
  long q = *d++;

  while (q*q <= p)
  {
    if (p%q == 0) return 0; 
    q += *d++;
  }
  return 1;
#endif
}

#if 0
static int
elladd(GEN *X1, GEN *X2)
{
  GEN lambda, *Y1 = X1+nbc, *Y2 = X2+nbc, *A = W+nbc+1;
  long i, av=avma;
  for (i=0; i<nbc; i++)
  {
    subiiz(X1[i], X2[i], A[i]); if (signe(A[i])<0) addiiz(A[i],N,A[i]);
    modiiz(mulii(A[i], W[i]), N, W[i+1]);
  }
  avma=av; /* if gl != N we have a factor */
  if (!inversemodulo(W[nbc], N, &gl)) return !egalii(N,gl);
  while (i--)
  {
    lambda = modii(mulii(subii(Y1[i], Y2[i]), mulii(gl, W[i])), N);
    modiiz(subii(sqri(lambda), addii(X2[i], X1[i])), N, X2[i]);
    modiiz(subii(mulii(lambda, subii(X1[i], X2[i])), Y1[i]), N, Y2[i]);
    modiiz(mulii(gl, A[i]), N, gl);
  }
  avma=av; return 0;
}
#else
/* variant with fewer assignments.  We use more stack space, and thus run
   a bit of a risk of overflowing it, but it's still bounded by a constant
   multiple of lgefint(N)*nbc, as it was in the old version --GN1998Jul02 */
static int
elladd(GEN *X1, GEN *X2)
{
  GEN lambda, *Y1 = X1+nbc, *Y2 = X2+nbc;
  GEN W[2*nbc], *A=W+nbc;	/* not the global W; W[0],A[0] never used */
  long i, av=avma, tetpil;
  /* W[0] = gun; */
  W[1] = /* A[0] =*/ subii(X1[0], X2[0]);
  for (i=1; i<nbc; i++)
  {
    A[i] = subii(X1[i], X2[i]);	/* don't waste time reducing mod N here */
    W[i+1] = modii(mulii(A[i], W[i]), N);
  }
  tetpil = avma;
  /* if gl != N we have a factor */
  if (!inversemodulo(W[nbc], N, &gl))
  { gl = gerepile(av,tetpil,gl); return !egalii(N,gl); }
  while (i--)			/* nbc times, actually */
  {
    lambda = modii(mulii(subii(Y1[i], Y2[i]),
			 i?mulii(gl, W[i]):gl), N);
    modiiz(subii(sqri(lambda), addii(X2[i], X1[i])), N, X2[i]);
    modiiz(subii(mulii(lambda, subii(X1[i], X2[i])), Y1[i]), N, Y2[i]);
    if (!i) break;
    gl = modii(mulii(gl, A[i]), N);
  }
  avma=av; return 0;
}
#endif

#if 0
static int
elldouble(GEN *X)
{
  GEN lambda,v, *Y = X+nbc;
  long i, av=avma;
  for (i=0; i<nbc; i++)
    modiiz(mulii(Y[i], W[i]), N, W[i+1]);
  avma = av; 
  if (!inversemodulo(W[nbc], N, &gl)) return !egalii(gl,N);
  while (i--)
  {
    lambda = modii(mulii(addsi(1, mulsi(3, mulii(X[i], X[i]))),
			 mulii(gl,W[i])), N);
    lambda = shifti(mpodd(lambda)? addii(lambda, N): lambda, -1);
    v = modii(subii(sqri(lambda), shifti(X[i],1)), N);
    modiiz(mulii(gl, Y[i]), N, gl);
    modiiz(subii(mulii(lambda, subii(X[i], v)), Y[i]), N, Y[i]);
    affii(v, X[i]);
  }
  avma = av; return 0;
}
#else
/* variant with fewer assignments.  Strangely, whereas this gains about
   3% on my P133 with elladd(), it makes hardly any difference here with
   elldouble() --GN */
static int
elldouble(GEN *X)
{
  GEN lambda,v, *Y = X+nbc;
  GEN W[nbc+1];			/* not the global W; W[0] never used */
  long i, av=avma, tetpil;
  /*W[0] = gun;*/ W[1] = Y[0];
  for (i=1; i<nbc; i++)
    W[i+1] = modii(mulii(Y[i], W[i]), N);
  tetpil = avma; 
  if (!inversemodulo(W[nbc], N, &gl))
  { gl = gerepile(av,tetpil,gl); return !egalii(N,gl); }
  while (i--)
  {
    lambda = modii(mulii(addsi(1, mulsi(3, sqri(X[i]))),
			 i?mulii(gl,W[i]):gl), N);
    lambda = shifti(mpodd(lambda)? addii(lambda, N): lambda, -1);
    v = modii(subii(sqri(lambda), shifti(X[i],1)), N);
    if (i) gl = modii(mulii(gl, Y[i]), N);
    modiiz(subii(mulii(lambda, subii(X[i], v)), Y[i]), N, Y[i]);
    affii(v, X[i]);
  }
  avma = av; return 0;
}
#endif

#if 0
static int
ellmult(long k, GEN *X, GEN *XAUX)
{
  if (k == 1)
  { 
    for (k = 2*nbc; k--; ) affii (X[k], XAUX[k]);
    return 0;
  }
  if (ellmult(k>>1, X, XAUX) || elldouble(X)) return !egalii(gl,N);
  return k&1? elladd(XAUX,X): 0;
}
#else
/* The following will never be called from outside with k==1 */
static int
ellmult(long k, GEN *X, GEN *XAUX)
{
  long k1;
  if (k == 1) return 0;
  if (k <= 3)
  { 
    for (k1 = 2*nbc; k1--; ) affii (X[k1], XAUX[k1]);
    if (elldouble(X)) return 1;
  }
  else
    if (ellmult(k>>1, X, XAUX) || elldouble(X)) return 1;
  return k&1? elladd(XAUX,X): 0;
}
#endif

/* ellfacteur() being re-tuned to be useful as a first stage before
   MPQS, especially for _large_ arguments, when insist is false.  Not
   yet final. --GN */
GEN
ellfacteur(GEN n, int insist)
{
  static long TB1[] =
    { 156,405,962,2240,4778,9004,18437,34155,66596,133297,
      247988,374990,649996,1170924,1967442,3276490,5249667 };
  static long TB1_for_stage[] =
    { /*50,*/ 65,84,114,160,211,289,456,729,1024,1448,2048,2896,
      4096,5792,8192,11586,16384,23172,32768 };
  /* 65536,131072,262144,524288,1048576,2097152,4194304 */
  long B1,B2,dsn,dsnmax,a,h,i,j,p,m,av,rep,spc, tf = lgefint(n);
  GEN *X,*XD,*XDD, res = cgeti(tf);

  av=avma; N = n;
  /* following moved up here so we don't waste time in the `decline' case */
  if (insist)
  {
    /* Somehow I don't think this had ever been right for 64-bit machines?
       It used to figure out dsn and hence B1, B2 based on number of words
       alone, which doesn't make sense if words can be twice as large on
       some machines than on others. --GN */
#ifdef LONG_IS_64BITS
    long tf32 = (tf-2) << 1;
#else
    long tf32 = tf - 2;
#endif
    dsn = (long)sqrt(9.6*tf32); dsnmax = (long)(4.8*tf32);
    if (dsnmax > 41) dsnmax = 41;
    if (dsn < 9) dsn = 9; else if (dsn > dsnmax) dsn = dsnmax;
    a = 1 + (nbc<<8);
  }
  else
  {
    long size = expi(n) + 1; dsn = size>>3;
    if (dsn > 34) dsn = 34; dsnmax = dsn+3; /* limit escalation */
    if (dsn < 19)		/* < 152 bits: decline the task */
    {
      if (DEBUGLEVEL>3)
      {
	fprintferr("ECM: number too small to justify this stage\n");
	flusherr();
      }
      avma = av; return NULL;
    }

    if (dsn < 23) rep = 2;	/* repetition count through main loop */
    else if (dsn < 27) rep = 3;
    else if (dsn < 31) rep = 4;
    else rep = 5;		/* may take hours... */

    /* it may be convenient to use disjoint sets of curves for the non-insist
       and insist phases;  moreover, repeated non-insistent calls acting on
       factors of the same original number should try to use fresh curves.
       The following achieves this */
    a = 1 + (nbc<<2)*(size & 0x3f);
  }
  spc = 258;			/* was 260 */
  W = 1 + (GEN*) cgeti(spc*nbc + 2);
  X   = W + 1;			/* was W  + 2*nbc + 1 */
  XD  = X  + 2*nbc;
  XDD = XD + 2*nbc;
  if (DEBUGLEVEL>3) { (void) timer2(); } /* clear timer */
  {
    GEN w = cgeti(spc*nbc*tf);
    W[0]=gun; m = evallg(tf) | evaltyp(t_INT);
    for (i = spc*nbc; i; i--)
    {
      *w = m; W[i] = w; w += tf; /* hack for: w = cgeti(tf) */
    }
  }
  tf -= 2;

  for(;;)
  {
    byteptr d = diffptr;

    for (i = 2*nbc; i--; ) affsi(a++, X[i]);
    if (insist) { B1 = TB1[dsn/2 - 4]; B2 = 50*B1; }
    else { B1 = TB1_for_stage[dsn - 19]; B2 = (7+dsn)*B1; }

    if (DEBUGLEVEL > 3)
    { 
      fprintferr("ECM: time = %6ld ms,\tdsn = %2d,\tB1 = %4d,\tB2 = %6d\n",
                 timer2(),dsn,B1,B2);
      flusherr(); 
    }
    p = *d++;
#if 1
    /* treat p=2 separately */
    for (m=2; m<B2; m<<=1)
      if (elldouble(X)) goto fin;
    p += *d++;
#endif
#if 1
    /* p=3:  usually the loop runs to the end, and the `times 9' appears in
       practice to be faster than `times 3, times 3' (three elldouble and one
       elladd instead of two and two and one extra copying) */
    for (m=1; m<B2; m*=9)
      if (ellmult(9,X,XD)) goto fin;
    p += *d++;
#endif
    for (; p < B1; p += *d++)
    {
      for (m=p; m<B2; m*=p)
	if (ellmult(p,X,XD)) goto fin;
      if (! *d) break;
    }
    while (p < B1)
    {
      do p += 2; while (! sisprime(p));
      for (m=p; m<B2; m*=p)
	if (ellmult(p,X,XD)) goto fin;
    }
    /* begin building table of even multiples of the current X, to be used
       later to go from `ellmult(some_prime_p,X,_)' to `ellmult(next_prime,
       X,_)' using a single addition */
    for (i=0; i < 2*nbc; i++) affii (X[i], XD[i]);
    if (elldouble(XD) || ellmult(p,X,XDD)) goto fin;
    for (i=0; i < 2*nbc; i++) affii (XD[i], XDD[i]);
    if (elldouble(XDD)) goto fin;
    /* however, be lazy and delay most of this until it's actually needed...
       This makes quite a noticeable difference (about 5%) --GN */
    h = (B1 >= 1024 ? 48 : 32);
    /* will never grow larger than 256 */
    if (DEBUGLEVEL>4)
      fprintferr("\tB2 stage, allowing for delta p <= %ld\n", h);
    for (i=4; i < h; i += 2)
    {
      for (j = (i-2)*nbc; j < i*nbc; j++) affii (XD[j], XDD[j]);
      if (elladd(XD, XD + i*nbc)) goto fin;
    }
    while (p < B2 && *d)
    {
      if (*d >= h)
      {
	long h1 = h + 8; while (h1 <= *d) h1 += 8;
	if (DEBUGLEVEL>4)
	  fprintferr("\tallowing for delta p <= %ld\n", h1);
	for (i=h; i < h1; i += 2)
	{
	  for (j = (i-2)*nbc; j < i*nbc; j++) affii (XD[j], XDD[j]);
	  if (elladd(XD, XD + i*nbc)) goto fin;
	}
	h = h1;
      }
      while (*d < h && p < B2 && *d)
      {
	p += *d; if (elladd(X + nbc * *d++, X)) goto fin;
      }
    }
    while (p < B2)
    {
      long oldp = p, dp;
      do p += 2; while (! sisprime(p));
      if ((dp = p - oldp) >= h)
      {
	long h1 = h + 8; while (h1 <= dp) h1 += 8;
	if (DEBUGLEVEL>4)
	  fprintferr("\tallowing for delta p <= %ld\n", h1);
	for (i=h; i < h1; i += 2)
	{
	  for (j = (i-2)*nbc; j < i*nbc; j++) affii (XD[j], XDD[j]);
	  if (elladd(XD, XD + i*nbc)) goto fin;
	}
	h = h1;
      }
      if (elladd(X + dp*nbc, X)) goto fin;
    }
    if (dsn < dsnmax) dsn += (insist ? 2 : 1);
#if 1
    if (!insist && !--rep)
    {
      if (DEBUGLEVEL > 3)
      { 
	fprintferr("ECM: time = %6ld ms,\tellfacteur giving up.\n",
		   timer2());
	flusherr(); 
      }
      avma = av; return NULL;
    }
#endif
  }
 fin:
#if 0
  affii(cmpii(sqri(gl),N) > 0 ? divii(N,gl): gl, res);
#else
  affii(gl, res);
#endif
  if (DEBUGLEVEL > 3)
  {
    fprintferr("ECM: time = %6ld ms,\tp = %6d,\n\tfound factor = %Z\n",
	       timer2(), p, (long)res);
    flusherr();
  }
  avma=av; return res;
}

/***********************************************************************/
/**                                                                   **/
/**                FACTORIZATION (Pollard-Brent rho)                  **/
/**  pollardbrent() returns a non trivial factor of n, assuming n is  **/
/**  composite and has no small prime divisor, or NULL if going on    **/
/**  would take more time than we want to spend.  GN1998Jun18-26      **/
/**                 (Cf. Algorithm 8.5.2 in ACiCNT)                   **/
/**                                                                   **/
/***********************************************************************/

/* Tuning parameter:  The overriding concern here is that for input up to
   64 bits long, we must not spend more than a very short time, because we
   risk slowing things down on average otherwise.  With the current tuning
   formula, we begin increasing our efforts somewhat at 49 bit input  (by
   an extra round for each bit at first),  and go up more and more rapidly
   after we pass 80 bits. */

#define tune_pb_min 28		/* even 30 seems too much */

/* We return NULL when we run out of time, or a single t_INT containing a
   nontrivial factor of n, or a vector of t_INTs, each triple of successive
   entries containing a factor, an exponent  (equal to un),  and a factor
   class  (NULL for unknown or zero for known composite),  matching the
   internal representation used by the ifac_*() routines below.  Repeated
   factors can arise and are legal;  the caller will be sorting the factors
   anyway. */
/* NB this function will shamelessly divide into n under certain conditions.
   It's the caller's responsibility to ensure that such behaviour is safe. */
GEN
pollardbrent(GEN n)
{
  GEN x=gdeux, x1=gdeux, y=gdeux, P=gun, g, g1=NULL;
  long tf = lgefint(n), size = 0, delta, retries = 0, av=avma;
  long c0, c, c1, k, k1, l;
  GEN res=cgeti(tf), res1;
  long av1=avma, lim=(av1+bot)>>1;

  if (DEBUGLEVEL>3) { (void) timer2(); } /* clear timer */

  if (tf >= 4)
    size = expi(n) + 1;
  else if (tf == 3)		/* try to keep purify happy...  */
    size = BITS_IN_LONG - bfffo(n[2]);
  if (size <= 32)
    c0 = 64;			/* amounts very nearly to `insist' */
  else if (size <= 48)
    c0 = tune_pb_min;
  else if (size <= 72)
    c0 = tune_pb_min + size - 48;
  else if (size <= 221)
    /* introduce nonlinear increase in effort,
     * kicking in around 80 bits */
    c0 = tune_pb_min + 24 + ((size-72)<<1) +
      ((size-73)>>1)*((size-70)>>3)*((size-56)>>4);
  /* 222 gives 14384 + tune_pb_min */
  else if (size < 65500)
    c0 = size<<6;		/* back to linear; ECM taking over */
  else
    c0 = 1<<22;
				/* improbable, but size _could_ be
				   very large indeed... */
  c = c0 << 4;			/* 16 iterations per round */

 PB_RETRY:
  k = 1; l = 1;
  /* trick to make a `random' choice determined by n.  Don't use x^2+0 or
     x^2-2, ever.  Don't use x^2-3 or x^2-7 with a starting value of 2.
     x^2+4, x^2+9 are affine conjugate to x^2+1, so don't use them either. */
  switch ((size + retries) & 7)
  {
  case 0:
    delta=1; break;
  case 1:
    delta=-1; break;
  case 2:
    delta=3; break;
  case 3:
    delta=5; break;
  case 4:
    delta=-5; break;
  case 5:
    delta=7; break;
  case 6:
    delta=11; break;
  case 7:
    delta=-11; break;
  }
  /* (the point being that when we get called again on a composite cofactor
   * of something we've already seen, we had better avoid the same delta)
   */
  if (DEBUGLEVEL>3)
  {
    if (!retries)
    {
      if (size < 1536)
	fprintferr("Rho: searching small factor of %ld-bit integer\n",
		   size);
      else
	fprintferr("Rho: searching small factor of %ld-word integer\n",
		   tf - 2);
      fprintferr("Rho: using X^2%+1ld for up to %ld rounds of 16 iterations\n",
		 delta, c0);
    }
    else
    {
      fprintferr("Rho: restarting for remaining rounds...\n");
      fprintferr("Rho: using X^2%+1ld for up to %ld rounds of 16 iterations\n",
		 delta, c >> 4);
    }
    flusherr();
  }
  
  for (;;)			/* terminated under the control of c */
  {
    /* we use the polynomial  x^2 + delta */
    x = modii(addsi(delta, sqri(x)), n);
    P = modii(mulii(P, subii(x1, x)), n);
    if (low_stack(lim, (av1+bot)>>1))
    {
      GEN *gsav[4];
      if(DEBUGMEM>1) err(warnmem,"[1]: pollardbrent");
      gsav[0]=&x; gsav[1]=&x1; gsav[2]=&y; gsav[3]=&P;
      gerepilemany(av1, gsav, 4);
    }
    if ((--c & 0xf)==0)		/* one round complete */
    {
      g = mppgcd(P, n);		/* see whether anything got caught */
      if (!is_pm1(g)) goto fin;
      y = icopy(x);
      P = gun;			/* not necessary, but saves one multiplication
				   per round */

      if (DEBUGLEVEL>3 && c>0 && ((c0 - (c>>4)) & 0xff)==0)
      {
	c1 = c0 - (c>>4);
	fprintferr("Rho: time = %6ld ms,\t%3ld rounds\n", /* always plural */
		   timer2(), c1);
	flusherr();
      }
    }
    if (c <= 0)
    {				/* getting bored */
      avma = av;
      if (DEBUGLEVEL>3)
      {
	fprintferr("Rho: time = %6ld ms,\tPollard-Brent giving up.\n",
		   timer2());
	flusherr();
      }
      return NULL;
    }

    if (--k) continue;		/* normal end of loop body */

    g = mppgcd(P, n);
    if (!is_pm1(g)) goto fin;
    P = gun;

    /* fast forward phase, doing l inner iterations without need to compute
       any gcds.  Check first whether it would take us beyond the alloted
       time.  Fast forward rounds count only half  (although they're taking
       more like 2/3 the time of normal rounds).  This to counteract the
       nuisance that all c0 between 4096 and 6144 would act exactly as
       4096;  with the halving trick only the range 4096..5120 collapses
       (similarly for all other powers of two). */

    if ((c-=(l>>1)) <= 0)
    {				/* got bored */
      avma = av;
      if (DEBUGLEVEL>3)
      {
	fprintferr("Rho: time = %6ld ms,\tPollard-Brent giving up.\n",
		   timer2());
	flusherr();
      }
      return NULL;
    }
    c &= ~0xf;			/* keep it on multiples of 16 */

    /* prepare for fast forward */
    x1 = icopy(x); k = l; l <<= 1;

    /* We don't show this for the first several (short) fast forward
       phases. */
    if (l > 8192 && DEBUGLEVEL > 3)
    {
      fprintferr("Rho: fast forward phase (%ld rounds of 32)...\n",
		 l >> 6);
      flusherr();
    }

    /* fast forward loop begins--- */
    for (k1=k; k1; k1--) {
      x = modii(addsi(delta, sqri(x)), n);
      P = modii(mulii(P, subii(x1, x)), n);
      if (low_stack(lim, (av1+bot)>>1))
      {
	GEN *gsav[4];
	if(DEBUGMEM>1) err(warnmem, "[2]: pollardbrent");
	gsav[0]=&x; gsav[1]=&x1; gsav[2]=&y; gsav[3]=&P;
	gerepilemany(av1, gsav, 4);
      }
    }
    /* ---that was it. */

    if (l > 8192 && DEBUGLEVEL > 3)
    {
      c1 = c0 - (c>>4);
      fprintferr("Rho: time = %6ld ms,\t%3ld rounds, back to normal mode\n",
		 timer2(), c1);
      flusherr();
    }

    /* one final thing before we're back in slow motion: */
    y = icopy(x);
  } /* forever */

  /* get here when a gcd accumulated in a round of 16 iterations, or in a
     fast forward phase followed by a normal round, was > 1 */
 fin:
  /* if g isn't n, and looks prime, we might as well return it at once */
  if  (!egalii(g,n))
  {
    if (miller(g,17))
    {
      affii(g, res);
      avma = av1;
      if (DEBUGLEVEL>3)
      {
	c1 = c0 - (c>>4);
	fprintferr("Rho: time = %6ld ms,\t%3ld round%s,\n",
		   timer2(), c1, (c1==1?"":"s"));
	fprintferr("\tfound factor = %Z\n",
		   (long)res);
	flusherr();
      }
      return res;
    }
    g1 = g;			/* known composite, keep on stack */
  }
  else g1 = n;			/* and work modulo g1 for backtracking */

  /* backtrack.  Here g1 is known composite */
  if (DEBUGLEVEL>3 && size > 256)
  {				/* all those gcds can take ages */
    fprintferr("Rho: hang on a second, we got something here...\n");
    flusherr();
  }
  for (;;)			/* must terminate */
  {
    y = modii(addsi(delta, sqri(y)), g1);
    g = mppgcd(subii(x1, y), g1);

    if (DEBUGLEVEL>3 && (--c & 0xf) == 0 && ((c0 - (c>>4)) & 0xff) == 0)
    {
      c1 = c0 - (c>>4);
      fprintferr("Rho: time = %6ld ms,\t%3ld rounds\n", /* always plural */
		 timer2(), c1);
      flusherr();
    }

    if (!is_pm1(g))
    {
      if (egalii(g,g1) || g1 == n)
      {
	if (egalii(g,g1) && g1 == n) /* out of luck */
	{
	  avma = av;
	  if (DEBUGLEVEL>3)
	  {
	    c1 = c0 - (c>>4);
	    fprintferr("Rho: time = %6ld ms,\t%3ld round%s,\n",
		       timer2(), c1, (c1==1?"":"s"));
	    fprintferr("\tPollard-Brent failed.\n");
	    flusherr();
	  }
	  if (++retries >= 4) return NULL;
	  goto PB_RETRY;
	}
	/* half lucky: we've split n, but g1 equals either g or n */
	mpdivis(n,g,res);
	if (DEBUGLEVEL>3)
	{
	  c1 = c0 - (c>>4);
	  fprintferr("Rho: time = %6ld ms,\t%3ld round%s,\n",
		     timer2(), c1, (c1==1?"":"s"));
	  fprintferr("\tfound %s factor = %Z\n",
		     (g1!=n ? "composite" : ""), (long)g);
	  flusherr();
	}
	res1 = cgetg(7, t_VEC);
	res1[1] = licopy(g);	/* factor */
	res1[2] = un;		/* exponent 1 */
	res1[3] = (g1!=n ? zero : (long)NULL);
				/* known composite when g1!=n */
	res1[4] = licopy(res);	/* cofactor */
	res1[5] = un;		/* exponent 1 */
	res1[6] = (long)NULL;	/* unknown */
	return gerepileupto(av,res1);
      }
      /* g < g1 < n : our lucky day -- now we've split g1, too */
      (void)mpdivis(n,g1,n);	/* must succeed */
      (void)mpdivis(g1,g,g1);	/* must succeed */
      /* now the original n is the product of the current g, g1, and n */
      if (DEBUGLEVEL>3)
      {
	c1 = c0 - (c>>4);
	fprintferr("Rho: time = %6ld ms,\t%3ld round%s,\n",
		   timer2(), c1, (c1==1?"":"s"));
	fprintferr("\tfound factors = %Z, %Z,\n\tand %Z\n",
		   (long)g, (long)g1, (long)n);
	flusherr();
      }
      res1 = cgetg(10, t_VEC);
      /* each of the three factors is of unknown status */
      res1[1] = licopy(g); res1[2] = un; res1[3] = (long)NULL;
      res1[4] = licopy(g1); res1[5] = un; res1[6] = (long)NULL;
      res1[7] = isonstack(n) ? licopy(n) : (long)n;
      res1[8] = un; res1[9] = (long)NULL;
      return gerepileupto(av,res1);
    }
    /* period not yet recovered -- keep backtracking */
    if (low_stack(lim, (av1+bot)>>1))
    {
      GEN *gsav[3];
      if(DEBUGMEM>1) err(warnmem,"[3]: pollardbrent");
      gsav[0]=&x1; gsav[1]=&y; gsav[2]=&g1;
      gerepilemany(av1, gsav, g1==n?2:3); /* leave g1 well alone when ==n */
    }
  } /* forever */
}

/***********************************************************************/
/**                                                                   **/
/**                      DETECTING ODD POWERS                         **/
/**  Factoring engines like MPQS which ultimately rely on computing   **/
/**  gcd(N, x^2-y^2) to find a nontrivial factor of N are fundamen-   **/
/**  tally incapable of splitting a proper power of an odd prime,     **/
/**  because of the cyclicity of the prime residue class group.  We   **/
/**  already have a square-detection function carrecomplet(), which   **/
/**  also returns the square root if appropriate.  Here's an analogue **/
/**  for cubes, fifth and 7th powers.  11th powers are a non-issue so **/
/**  long as mpqs() gives up beyond 100 decimal digits  (since ECM    **/
/**  easily find a 10-digit prime factor of a 100-digit number).      **/
/**  GN1998Jun28                                                      **/
/**                                                                   **/
/***********************************************************************/

/* Use a multistage sieve.  First stages work mod 211, 209, 61, 203;
   if the argument is larger than a word, we first reduce mod the product
   of these and then take the remainder apart.  Second stages use 117,
   31, 43, 71 in this order.  Moduli which are no longer interesting are
   skipped.  Everything is encoded in a single table of 106 24-bit masks.
   We only need the first half of the residues.  Three bits per modulus
   indicate which residues are 7th (bit 2), 5th (bit 1) powers or cubes
   (bit 0);  the eight moduli above are assigned right-to-left.  The table
   will err on the side of safety if one of the moduli divides the number
   to be tested, but as this leads to inefficiency it should still be
   avoided. */

static ulong powersmod[106] = {
  077777777ul,			/* 0 */
  077777777ul,			/* 1 */
  013562440ul,			/* 2 */
  012462540ul,			/* 3 */
  013562440ul,			/* 4 */
  052662441ul,			/* 5 */
  016663440ul,			/* 6 */
  016463450ul,			/* 7 */
  013573551ul,			/* 8 */
  012462540ul,			/* 9 */
  012462464ul,			/* 10 */
  013462771ul,			/* 11 */
  012466473ul,			/* 12 */
  012463641ul,			/* 13 */
  052463646ul,			/* 14 */
  012563446ul,			/* 15 */
  013762440ul,			/* 16 */
  052766440ul,			/* 17 */
  012772451ul,			/* 18 */
  012762454ul,			/* 19 */
  032763550ul,			/* 20 */
  013763664ul,			/* 21 */
  017763460ul,			/* 22 */
  037762565ul,			/* 23 */
  017762540ul,			/* 24 */
  057762441ul,			/* 25 */
  037772452ul,			/* 26 */
  017773551ul,			/* 27 */
  017767541ul,			/* 28 */
  017767640ul,			/* 29 */
  037766450ul,			/* 30 */
  017762752ul,			/* 31 */
  037762762ul,			/* 32 */
  017762742ul,			/* 33 */
  037763762ul,			/* 34 */
  017763740ul,			/* 35 */
  077763740ul,			/* 36 */
  077762750ul,			/* 37 */
  077762752ul,			/* 38 */
  077762750ul,			/* 39 */
  077762743ul,			/* 40 */
  077767740ul,			/* 41 */
  077763741ul,			/* 42 */
  077763762ul,			/* 43 */
  077772760ul,			/* 44 */
  077762770ul,			/* 45 */
  077766750ul,			/* 46 */
  077762740ul,			/* 47 */
  077763740ul,			/* 48 */
  077763750ul,			/* 49 */
  077763752ul,			/* 50 */
  077762740ul,			/* 51 */
  077762740ul,			/* 52 */
  077772740ul,			/* 53 */
  077762762ul,			/* 54 */
  077763765ul,			/* 55 */
  077763770ul,			/* 56 */
  077767750ul,			/* 57 */
  077766753ul,			/* 58 */
  077776740ul,			/* 59 */
  077772741ul,			/* 60 */
  077772744ul,			/* 61 */
  077773740ul,			/* 62 */
  077773743ul,			/* 63 */
  077773751ul,			/* 64 */
  077772771ul,			/* 65 */
  077772760ul,			/* 66 */
  077772763ul,			/* 67 */
  077772751ul,			/* 68 */
  077773750ul,			/* 69 */
  077777740ul,			/* 70 */
  077773745ul,			/* 71 */
  077772740ul,			/* 72 */
  077772742ul,			/* 73 */
  077772744ul,			/* 74 */
  077776750ul,			/* 75 */
  077773771ul,			/* 76 */
  077773774ul,			/* 77 */
  077773760ul,			/* 78 */
  077772741ul,			/* 79 */
  077772740ul,			/* 80 */
  077772740ul,			/* 81 */
  077772741ul,			/* 82 */
  077773754ul,			/* 83 */
  077773750ul,			/* 84 */
  077773740ul,			/* 85 */
  077776741ul,			/* 86 */
  077776771ul,			/* 87 */
  077776773ul,			/* 88 */
  077772761ul,			/* 89 */
  077773741ul,			/* 90 */
  077773740ul,			/* 91 */
  077773740ul,			/* 92 */
  077772740ul,			/* 93 */
  077772752ul,			/* 94 */
  077772750ul,			/* 95 */
  077772751ul,			/* 96 */
  077773741ul,			/* 97 */
  077773761ul,			/* 98 */
  077777760ul,			/* 99 */
  077772765ul,			/* 100 */
  077772742ul,			/* 101 */
  077777751ul,			/* 102 */
  077777750ul,			/* 103 */
  077777745ul,			/* 104 */
  077777770ul			/* 105 */
};

/* Returns 3, 5, or 7 if x is a cube (but not a 5th or 7th power),  a 5th
   power (but not a 7th),  or a 7th power, and in this case creates the
   base on the stack and assigns its address to *pt.  Otherwise returns 0.
   x must be of type t_INT and nonzero;  this is not checked.  The *mask
   argument tells us which things to check -- bit 0: 3rd, bit 1: 5th,
   bit 2: 7th pwr;  set a bit to have the corresponding power examined --
   and is updated appropriately for a possible follow-up call */

long				/* no longer static -- used in mpqs.c */
is_odd_power(GEN x, GEN *pt, long *mask)
{
  long av=avma, tetpil, lgx=lgefint(x), exponent=0, residue, resbyte;
  GEN y;

  *mask &= 7;			/* paranoia */
  if (!*mask) return 0;		/* useful when running in a loop */
  if (signe(x) < 0) x=absi(x);

  if (DEBUGLEVEL>4)
  {
    fprintferr("OddPwrs: is %Z\n\t...a", x);
    if (*mask&1) fprintferr(" 3rd%s",
			    (*mask==7?",":(*mask!=1?" or":"")));
    if (*mask&2) fprintferr(" 5th%s",
			    (*mask==7?", or":(*mask&4?" or":"")));
    if (*mask&4) fprintferr(" 7th");
    fprintferr(" power?\n");
  }
  if (lgx > 3) residue = smodis(x, 211*209*61*203);
  else residue = x[2];

  resbyte=residue%211; if (resbyte > 105) resbyte = 211 - resbyte;
  *mask &= powersmod[resbyte];
  if (DEBUGLEVEL>4)
  {
    fprintferr("\tmodulo: resid. (remaining possibilities)\n");
    fprintferr("\t   211:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
	       resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
  }
  if (!*mask) { avma=av; return 0; }

  if (*mask & 3)
  {
    resbyte=residue%209; if (resbyte > 104) resbyte = 209 - resbyte;
    *mask &= (powersmod[resbyte] >> 3);
    if (DEBUGLEVEL>4)
      fprintferr("\t   209:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }
  if (*mask & 3)
  {
    resbyte=residue%61; if (resbyte > 30) resbyte = 61 - resbyte;
    *mask &= (powersmod[resbyte] >> 6);
    if (DEBUGLEVEL>4)
      fprintferr("\t    61:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }
  if (*mask & 5)
  {
    resbyte=residue%203; if (resbyte > 101) resbyte = 203 - resbyte;
    *mask &= (powersmod[resbyte] >> 9);
    if (DEBUGLEVEL>4)
      fprintferr("\t   203:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }

  if (lgx > 3) residue = smodis(x, 117*31*43*71);
  else residue = x[2];

  if (*mask & 1)
  {
    resbyte=residue%117; if (resbyte > 58) resbyte = 117 - resbyte;
    *mask &= (powersmod[resbyte] >> 12);
    if (DEBUGLEVEL>4)
      fprintferr("\t   117:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }
  if (*mask & 3)
  {
    resbyte=residue%31; if (resbyte > 15) resbyte = 31 - resbyte;
    *mask &= (powersmod[resbyte] >> 15);
    if (DEBUGLEVEL>4)
      fprintferr("\t    31:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }
  if (*mask & 5)
  {
    resbyte=residue%43; if (resbyte > 21) resbyte = 43 - resbyte;
    *mask &= (powersmod[resbyte] >> 18);
    if (DEBUGLEVEL>4)
      fprintferr("\t    43:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }
  if (*mask & 6)
  {
    resbyte=residue%71; if (resbyte > 35) resbyte = 71 - resbyte;
    *mask &= (powersmod[resbyte] >> 21);
    if (DEBUGLEVEL>4)
      fprintferr("\t    71:  %3ld   (3rd %ld, 5th %ld, 7th %ld)\n",
		 resbyte, *mask&1, (*mask>>1)&1, (*mask>>2)&1);
    if (!*mask) { avma=av; return 0; }
  }

  /* priority to higher powers -- if we have a 21st, it'll be easier to
     rediscover that its 7th root is a cube than that its cube root is
     a 7th power */
  if ((resbyte = *mask & 4))	/* assignment */
    exponent = 7;
  else if ((resbyte = *mask & 2))
    exponent = 5;
  else
    { resbyte = 1; exponent = 3; }
  /* leave that mask bit on for the moment, we might need it for a
     subsequent call */

  /* precision in the following is one extra significant word (overkill) */
  y=ground(gpow(x, ginv(stoi(exponent)), lgx - 1));
  if (!egalii(gpowgs(y, exponent), x))
  {
    if (DEBUGLEVEL>4)
    {
      if (exponent == 3)
	fprintferr("\tBut it nevertheless wasn't a cube.\n");
      else
	fprintferr("\tBut it nevertheless wasn't a %ldth power.\n",
		   exponent);
    }
    *mask &= ~resbyte;		/* _now_ turn the bit off */
    avma=av; return 0;
  }
  /* caller (ifac_crack() below) will report the final result if it was
     a pure power, so no further diagnostics here */

  tetpil=avma;
  if (!pt) { avma=av; return exponent; } /* this branch not used */
  *pt=gerepile(av,tetpil,icopy(y));
  return exponent;
}

/***********************************************************************/
/**                                                                   **/
/**                FACTORIZATION  (master iteration)                  **/
/**      Driver for the various methods of finding large factors      **/
/**      (after trial division has cast out the very small ones).     **/
/**                        GN1998Jun24--30                            **/
/**                                                                   **/
/***********************************************************************/

/**  Direct use:
 **  ifac_start()  registers a number  (without prime factors < 100)
 **    with the iterative factorizer, and also registers whether or
 **    not we should terminate early if we find that the number is
 **    not squarefree, and a hint about which method(s) to use.  This
 **    must always be called first.  The input _must_ have been checked
 **    to be composite by the caller.  The routine immediately tries
 **    to decompose it nontrivially into a product of two factors,
 **    except in squarefreeness (`Moebius') mode.
 **  ifac_primary_factor()  returns a prime divisor  (not necessarily
 **    the smallest)  and the corresponding exponent. */

/**  Encapsulated user interface:
 **  ifac_decomp()  does the right thing for auxdecomp()  (put a succession
 **    of prime divisor / exponent pairs onto the stack, not necessarily
 **    sorted, although in practice they will tend not to be too far from
 **    the correct order).
 **
 **  For each of the additive/multiplicative arithmetic functions, there is
 **  a `contributor' below, to be called on any large composite cofactor
 **  left over after trial division by small primes, whose result can then
 **  be added to or multiplied with whatever we already have:
 **  ifac_moebius()  ifac_issquarefree()  ifac_totient()  ifac_omega()
 **  ifac_bigomega()  ifac_numdiv()  ifac_sumdiv()  ifac_sumdivk() */

/* We never test whether the input number is prime or composite, since
   presumably it will have come out of the small factors finder stage
   (which doesn't really exist yet but which will test the left-over
   cofactor for primality once it does). */

/* The data structure in which we preserve whatever we know at any given
   time about our number N is kept on the PARI stack, and updated as needed.
   This makes the machinery re-entrant  (you can have more than one fac-
   torization using ifac_start()/ifac_primary_factor() in progress simul-
   taneously so long as you preserve the GEN across garbage collections),
   and which avoids memory leaks when a lengthy factorization is interrupted.
   We also make an effort to keep the whole affair connected, and the parent
   object will always be older than its children.  This may in rare cases
   lead to some extra copying around, and knowing what is garbage at any
   given time is not entirely trivial.  See below for examples how to do
   it right.  (Connectedness can be destroyed if callers of ifac_main()
   create other stuff on the stack in between calls.  This is harmless
   as long as ifac_realloc() is used to re-create a connected object at
   the head of the stack just before collecting garbage.) */

/* Note that a PARI integer can have hundreds of millions of distinct prime
   factors larger than 2^16, given enough memory.  And since there's no
   guarantee that we will find factors in order of increasing size, we must
   be prepared to drag a very large amount of data around  (although this
   will _very_ rarely happen for random input!).  So we start with a small
   structure and extend it when necessary. */

/* The idea of data structure and algorithm is:
   Let N0 be whatever is currently left of N after dividing off all the
   prime powers we have already returned to the caller.  Then we maintain
   N0 as a product
   (1)   N0 = \prod_i P_i^{e_i} * \prod_j Q_j^{f_j} * \prod_k C_k^{g_k}
   where the P_i and Q_j are distinct primes, each C_k is known composite,
   none of the P_i divides any C_k, and we also know the total ordering
   of all the P_i, Q_j and C_k  (in particular, we will never try to divide
   a C_k by a larger Q_j).  Some of the C_k may have common factors, although
   this will not often be the case. */

/* Caveat implementor:  Taking gcds among C_k's is very likely to cost at
   least as much time as dividing off any primes as we find them, and book-
   keeping would be a nightmare  (since D=gcd(C_1,C_2) can still have common
   factors with both C_1/D and C_2/D, and so on...). */

/* At startup, we just initialize the structure to
   (2)        N = C_1^1   (composite). */

/* Whenever ifac_primary_factor() or ifac_decomp()  (or, mutatis mutandis,
   one of the three arithmetic user interface routines)  needs a primary
   factor, and the smallest thing in our list is P_1, we return that and
   its exponent, and remove it from our list.
   (When nothing is left, we return a sentinel value -- gun.  And in Moebius
   mode, when we see something with exponent > 1, whether prime or composite,
   we yell at our caller by returning gzero or 0, depending on the function).
   In all other cases, ifac_main() iterates the following steps until we have
   a P_1 in the smallest position. */

/* When the smallest item is C_1  (as it is initially):
   (3.1) Crack C_1 into a nontrivial product  U_1 * U_2  by whatever method
   comes to mind for this size.  (U for `unknown'.)  Cracking will detect
   squares  (and biquadrates etc),  and it may detect odd powers, so we
   might instead see a power of some U_1 here, or even something of the form
   U_1^k*U_2^k.  (Of course the exponent already attached to C_1 is taken
   into account in the following.)
   (3.2) If we have U_1*U_2, sort the two factors;  convert to U_1^2 if they
   happen to be equal  (which they shouldn't -- squares should have been
   caught at the preceding stage).  Note that U_1 and  (if it exists)  U_2
   are automatically smaller than anything else in our list.
   (3.3) Check U_1  (and U_2)  for primality, and flag them accordingly.
   (3.4) Iterate. */

/* When the smallest item is Q_1:
   This is the potentially unpleasant case.  The idea is to go through the
   entire list and try to divide Q_1 off each of the current C_k's, which
   will usually fail, but may succeed several times.  When a division was
   successful, the corresponding C_k is removed from our list, and the co-
   factor becomes a U_l for the moment unless it is 1  (which happens when
   C_k was a power of Q_1).  When we're through we upgrade Q_1 to P_1 status,
   and then do a primality check on each U_l and sort it back into the list
   either as a Q_j or as a C_k.  If during the insertion sort we discover
   that some U_l equals some P_i or Q_j or C_k we already have, we just add
   U_l's exponent to that of its twin.  (The sorting should therefore happen
   before the primality test).
   Note that this may produce one or more elements smaller than the P_1
   we just confirmed, so we may have to repeat the iteration. */

/* There's a little trick that avoids some Q_1 instances.  Just after we do
   a sweep to classify all current unknowns as either composites or primes,
   we do another downward sweep beginning with the largest current factor
   and stopping just above the largest current composite.  Every Q_j we
   pass is turned into a P_i.  (Different primes are automatically coprime
   among each other, and primes tend not to divide smaller composites.) */

/* (We have no use for comparing the square of a prime to N0.  Normally
   we will get called after casting out only the smallest primes, and
   since we cannot guarantee that we see the large prime factors in as-
   cending order, we cannot stop when we find one larger than sqrt(N0).) */

/* Data structure:  We keep everything in a single t_VEC of t_INTs.  The
   first component records whether we're doing full (NULL) or Moebius (un)
   factorization;  in the latter case many subroutines return a sentinel
   value as soon as they spot an exponent > 1.  The second component records
   the hint from factorint()'s optional flag, for use by ifac_crack().
   The remaining components  (initially 15)  are used in groups of three:
   a GEN pointer at the t_INT value of the factor, a pointer at the t_INT
   exponent  (usually gun or gdeux so we don't clutter up the stack too
   much),  and another t_INT GEN pointer to record the class of the factor:
   NULL for unknown, zero for known composite C_k, un for known prime Q_j
   awaiting trial division, and deux for finished prime P_i. */

/* When during the division stage we re-sort a C_k-turned-U_l to a lower
   position, we rotate any intervening material upward towards its old
   slot.  When a C_k was divided down to 1, its slot is left empty at
   first;  similarly when the re-sorting detects a repeated factor.
   After the sorting phase, we de-fragment the list and squeeze all the
   occupied slots together to the high end, so that ifac_crack() has room
   for new factors.  When this doesn't suffice, we abandon the current
   vector and allocate a somewhat larger one, defragmenting again during
   copying. */

/* (For internal use, note that all exponents will fit into C longs, given
   PARI's lgefint field size.  When we work with them, we sometimes read
   out the GEN pointer, and sometimes do an itos, whatever is more con-
   venient for the task at hand.) */


/*** Overview and forward declarations: ***/

/* The `*where' argument in the following points into *partial at the
   first of the three fields of the first occupied slot.  It's there
   because the caller would already know where `here' is, so we don't
   want to search for it again, although it wouldn't take much time.
   On the other hand, we do not preserve this from one user-interface
   call to the next. */

static GEN
ifac_find(GEN *partial, GEN *where);
/* Return GEN pointing at the first nonempty slot strictly behind the
   current *where, or NULL if such doesn't exist.  Can be used to skip
   a range of vacant slots, or to initialize *where in the first place
   (pass partial in both args).  Does not modify its argument pointers. */

void
ifac_realloc(GEN *partial, GEN *where, long new_lg);
/* Move to a larger main vector, updating *where if it points into it.
   Certainly updates *partial.  Can be used as a specialized gcopy before
   a gerepileupto()/gerepilemanysp()  (pass 0 as the new length).
   Normally, one would pass new_lg=1 to let this function guess the
   new size.  To be used sparingly. */

static long
ifac_crack(GEN *partial, GEN *where);
/* Split the first (composite) entry.  There _must_ already be room for
   another factor below *where, and *where will be updated.  Factor and
   cofactor will be inserted in the correct order, updating *where, or
   factor^k will be inserted if such should be the case  (leaving *where
   unchanged).  The factor or factors will be set to unknown, and inherit
   the exponent  (or a multiple thereof)  of its/their ancestor.  Returns
   number of factors written into the structure  (normally 2, but 1 if a
   factor equalled its cofactor, and may be more than 1 if a factoring
   engine returned a vector of factors instead of a single factor).  Can
   reallocate the data structure in the vector-of-factors case  (but not
   in the more common single-factor case) */

static long
ifac_insert_multiplet(GEN *partial, GEN *where, GEN facvec);
/* Gets called to complete ifac_crack()'s job when a factoring engine
   splits the current factor into a product of three or more new factors.
   Makes room for them if necessary, sorts them, gives them the right
   exponents and class etc.  Also returns the number of factors actually
   written, which may be less than the number of components in facvec
   if there are duplicates.--- Vectors of factors  (cf pollardbrent()
   above)  actually contain `slots' of three GENs per factor with the
   three fields being interpreted exactly as in our partial factorization
   data structure.  Thus `engines' can tell us what they already happen to
   know about factors being prime or composite and/or appearing to a power
   larger than the first */

static long
ifac_divide(GEN *partial, GEN *where);
/* Divide all current composites by first  (prime, class Q)  entry, updating
   its exponent, and turning it into a finished prime  (class P).  Return 1
   if any such divisions succeeded  (in Moebius mode, the update may then
   not have been completed),  or 0 if none of them succeeded.  Doesn't
   modify *where. */

static long
ifac_sort_one(GEN *partial, GEN *where, GEN washere);
/* re-sort one  (typically unknown)  entry from washere to a new position,
   rotating intervening entries upward to fill the vacant space.  It may
   happen (rarely) that the new position is the same as the old one, or
   that the new value of the entry coincides with a value already occupying
   a lower slot, in which latter case we just add exponents  (and use the
   `more known' class, and return 1 immediately when in Moebius mode).
   The slots between *where and washere must be in sorted order, so a
   sweep using this to re-sort several unknowns must proceed upward  (see
   ifac_resort() below).  Return 1 if we see an exponent > 1  (in Moebius
   mode without completing the update),  0 otherwise. */

static long
ifac_resort(GEN *partial, GEN *where);
/* sort all current unknowns downward to where they belong.  Sweeps
   in the upward direction.  Not needed after ifac_crack(), only when
   ifac_divide() returned true.  May update *where.  Returns 1 when an
   ifac_sort_one() call does so to indicate a repeated factor, or 0 if
   any and all such calls returned 0 */

static void
ifac_defrag(GEN *partial, GEN *where);
/* defragment: collect and squeeze out any unoccupied slots above *where
   during a downward sweep.  Unoccupied slots arise when a composite factor
   dissolves completely whilst dividing off a prime, or when ifac_resort()
   spots a coincidence and merges two factors.  *where will be updated */

static void
ifac_whoiswho(GEN *partial, GEN *where, long after_crack);
/* determine primality or compositeness of all current unknowns, and set
   class Q primes to finished (class P) if everything larger is already
   known to be prime.  When after_crack is nonnegative, only look at the
   first after_crack things in the list (do nothing when it's zero) */

static GEN
ifac_main(GEN *partial);
/* main loop:  iterate until smallest entry is a finished prime;  returns
   a `where' pointer, or gun if nothing left, or gzero in Moebius mode if
   we aren't squarefree */

/* NB In the most common cases, control flows from the user interface to
   ifac_main() and then to a succession of ifac_crack()s and ifac_divide()s,
   with (typically) none of the latter finding anything. */

/** user interface: **/
/* return initial data structure, see ifac_crack() below for semantics
   of the hint argument */
GEN
ifac_start(GEN n, long moebius, long hint);

/* run main loop until primary factor is found, return the prime and
   assign the exponent.  If nothing left, return gun and set exponent
   to 0;  if in Moebius mode and a square factor is discovered, return
   gzero and set exponent to 0 */
GEN
ifac_primary_factor(GEN *partial, long *exponent);

/* call ifac_start() and run main loop until factorization is complete,
   accumulating prime / exponent pairs on the PARI stack to be picked up
   by aux_end().  Return number of distinct primes found */
long
ifac_decomp(GEN n, long hint);

/* completely encapsulated functions;  these call ifac_start() themselves,
   and ensure proper stack housekeeping etc.  Call them on any large
   composite left over after trial division, and multiply/add the result
   onto whatever you already have from the small factors.  Don't call
   them on large primes;  they will run into trouble */
long
ifac_moebius(GEN n, long hint);

long
ifac_issquarefree(GEN n, long hint);

long
ifac_omega(GEN n, long hint);

long
ifac_bigomega(GEN n, long hint);

GEN
ifac_totient(GEN n, long hint);	/* for gp's eulerphi() */

GEN
ifac_numdiv(GEN n, long hint);

GEN
ifac_sumdiv(GEN n, long hint);

GEN
ifac_sumdivk(GEN n, long k, long hint);

/*** implementation ***/

#define ifac_initial_length 24	/* codeword, moebius flag, hint, 7 slots */
/* (more than enough in most cases -- a 512-bit product of distinct 8-bit
   primes needs at most 7 slots at a time) */

GEN
ifac_start(GEN n, long moebius, long hint)
{
  GEN part, here;

  if (typ(n) != t_INT) err(typeer, "ifac_start");
  if (signe(n) == 0)
    err(talker, "factoring 0 in ifac_start");

  part = cgetg(ifac_initial_length, t_VEC);
  here = part + ifac_initial_length;
  part[1] = moebius? un : (long)NULL;
  switch(hint)
  {
  case 0:
    part[2] = zero; break;
  case 1:
    part[2] = un; break;
  case 2:
    part[2] = deux; break;
  default:
    part[2] = (long)stoi(hint);
  }
  if (isonstack(n))
    n = absi(n);
  /* make copy, because we'll later want to mpdivis() into it in place.
     If it's not on stack, then we assume it is a clone made for us by
     auxdecomp0(), and we assume the sign has already been set positive */
  /* fill first slot at the top end */
  *--here = zero;		/* initially composite */
  *--here = un;			/* initial exponent 1 */
  *--here = (long) n;
  /* and NULL out the remaining slots */
  while (here > part + 3) *--here = (long)NULL;
  return part;
}

static GEN
ifac_find(GEN *partial, GEN *where)
{
  long lgp = lg(*partial);
  GEN end = *partial + lgp;
  GEN scan = *where + 3;

  if (DEBUGLEVEL > 4)
  {
    if (!*partial || typ(*partial) != t_VEC)
      err(typeer, "ifac_find");
    if (lg(*partial) < ifac_initial_length)
      err(talker, "partial impossibly short in ifac_find");
    if (!(*where) ||
	*where > *partial + lgp - 3 ||
        *where < *partial)	/* sic */
      err(talker, "`*where\' out of bounds in ifac_find");
  }
  while (scan < end && !*scan) scan += 3;
  /* paranoia -- check completely NULLed ? nope -- we never inspect the
     exponent field for deciding whether a slot is empty or occupied */
  if (scan < end)
  {
    if (DEBUGLEVEL > 4)
    {
      if (!scan[1])
	err(talker, "factor has NULL exponent in ifac_find");
    }
    return scan;
  }
  return NULL;
}

/* simple defragmenter */
static void
ifac_defrag(GEN *partial, GEN *where)
{
  long lgp = lg(*partial);
  GEN scan_new = *partial + lgp - 3, scan_old = scan_new;

  while (scan_old >= *where)
  {
    if (*scan_old)		/* slot occupied? */
    {
      if (scan_old < scan_new)
      {
	scan_new[2] = scan_old[2];
	scan_new[1] = scan_old[1];
	*scan_new = *scan_old;
      }
      scan_new -= 3;		/* point at next slot to be written */
    }
    scan_old -= 3;
  }
  scan_new += 3;		/* back up to last slot written */
  *where = scan_new;
  while (scan_new > *partial + 3)
    *--scan_new = (long)NULL;	/* erase junk */
}

/* and complex version combined with reallocation.  If new_lg is 0, we
   use the old length, so this acts just like gcopy except that the where
   pointer is carried along;  if it is 1, we make an educated guess.
   Exception:  If new_lg is 0, the vector is full to the brim, and the
   first entry is composite, we make it longer to avoid being called again
   a microsecond later  (at significant cost).
   It is safe to call this with NULL for the where argument;  if it doesn't
   point anywhere within the old structure, it will be left alone */
void
ifac_realloc(GEN *partial, GEN *where, long new_lg)
{
  long old_lg = lg(*partial);
  GEN newpart, scan_new, scan_old;

  if (DEBUGLEVEL > 4)		/* none of these should ever happen */
  {
    if (!*partial || typ(*partial) != t_VEC)
      err(typeer, "ifac_realloc");
    if (lg(*partial) < ifac_initial_length)
      err(talker, "partial impossibly short in ifac_realloc");
  }

  if (new_lg == 1)
    new_lg = 2*old_lg - 6;	/* from 7 slots to 13 to 25... */
  else if (new_lg <= old_lg)	/* includes case new_lg == 0 */
  {
    new_lg = old_lg;
    if ((*partial)[3] &&	/* structure full */
	((*partial)[5]==zero || (*partial)[5]==(long)NULL))
				/* and first entry composite or unknown */
      new_lg += 6;		/* give it a little more breathing space */
  }
  newpart = cgetg(new_lg, t_VEC);
  if (DEBUGMEM >= 3)
  {
    fprintferr("IFAC: new partial factorization structure (%ld slots)\n",
	       (new_lg - 3)/3);
    flusherr();
  }
  newpart[1] = (*partial)[1];	/* moebius */
  newpart[2] = (*partial)[2];	/* hint */
  /* downward sweep through the old *partial, picking up where1 and carry-
     ing it over if and when we pass it.  (This will only be useful if
     it pointed at a non-empty slot.)  Factors are licopy()d so that we
     again have a nice object  (parent older than children, connected),
     except the one factor that may still be living in a clone where n
     originally was;  exponents are similarly copied if they aren't global
     constants;  class-of-factor fields are always global constants so we
     need only copy them as pointers.  Caller may then do a gerepileupto()
     or a gerepilemanysp() */
  scan_new = newpart + new_lg - 3;
  scan_old = *partial + old_lg - 3;
  for (; scan_old > *partial + 2; scan_old -= 3)
  {
    if (*where == scan_old) *where = scan_new;
    if (!*scan_old) continue;	/* skip empty slots */

    *scan_new =
      isonstack((GEN)(*scan_old)) ?
	licopy((GEN)(*scan_old)) : *scan_old;
    scan_new[1] =
      isonstack((GEN)(scan_old[1])) ?
	licopy((GEN)(scan_old[1])) : scan_old[1];
    scan_new[2] = scan_old[2];
    scan_new -= 3;
  }
  scan_new += 3;		/* back up to last slot written */
  while (scan_new > newpart + 3)
    *--scan_new = (long)NULL;
  *partial = newpart;
}

#define moebius_mode ((*partial)[1])

/* Bubble-sort-of-thing sort.  Won't be exercised frequently,
   so this is ok. */
static long
ifac_sort_one(GEN *partial, GEN *where, GEN washere)
{
  GEN scan = washere - 3;
  GEN value, exponent, class0, class1;
  long cmp_res;

  if (DEBUGLEVEL > 4)		/* none of these should ever happen */
  {
    long lgp;
    if (!*partial || typ(*partial) != t_VEC)
      err(typeer, "ifac_sort_one");
    if ((lgp = lg(*partial)) < ifac_initial_length)
      err(talker, "partial impossibly short in ifac_sort_one");
    if (!(*where) ||
	*where < *partial + 3 ||
	*where > *partial + lgp - 3)
      err(talker, "`*where\' out of bounds in ifac_sort_one");
    if (!washere ||
	washere < *where ||
	washere > *partial + lgp - 3)
      err(talker, "`washere\' out of bounds in ifac_sort_one");
  }
  value = (GEN)(*washere);
  exponent = (GEN)(washere[1]);
  if (exponent != gun && moebius_mode && cmpsi(1,exponent) < 0)
    return 1;			/* should have been detected by caller */
  class0 = (GEN)(washere[2]);

  if (scan < *where) return 0;	/* nothing to do, washere==*where */

  cmp_res = -1;			/* sentinel */
  while (scan >= *where)	/* therefore at least once */
  {
    if (*scan)			/* current slot nonempty */
    {
      /* check against where */
      cmp_res = cmpii(value, (GEN)(*scan));
      if (cmp_res >= 0) break;	/* have found where to stop */
    }
    /* copy current slot upward by one position and move pointers down */
    scan[5] = scan[2];
    scan[4] = scan[1];
    scan[3] = *scan;
    scan -= 3;
  }
  scan += 3;
  /* at this point there are the following possibilities:
     (*) cmp_res == -1.  Either value is less than that at *where, or for
     some reason *where was pointing at one or more vacant slots and any
     factors we saw en route were larger than value.  At any rate,
     scan == *where now, and scan is pointing at an empty slot, into
     which we'll stash our entry.
     (*) cmp_res == 0.  The entry at scan-3 is the one, we compare class0
     fields and add exponents, and put it all into the vacated scan slot,
     NULLing the one at scan-3  (and possibly updating *where).
     (*) cmp_res == 1.  The slot at scan is the one to store our entry
     into. */
  if (cmp_res != 0)
  {
    if (cmp_res < 0 && scan != *where)
      err(talker, "misaligned partial detected in ifac_sort_one");
    *scan = (long)value;
    scan[1] = (long)exponent;
    scan[2] = (long)class0;
    return 0;
  }
  /* case cmp_res == 0: repeated factor detected */
  if (DEBUGLEVEL >= 4)
  {
    fprintferr("IFAC: repeated factor %Z\n\tdetected in ifac_sort_one\n",
	       (long)value);
    flusherr();
  }
  if (moebius_mode) return 1;	/* not squarefree */
  /* if old class0 was composite and new is prime, or vice versa,
     complain  (and if one class0 was unknown and the other wasn't,
     use the known one) */
  class1 = (GEN)(scan[-1]);
  if (class0)			/* should never be used */
  {
    if(class1)
    {
      if (class0 == gzero && class1 != gzero)
	err(talker, "composite equals prime in ifac_sort_one");
      else if (class0 != gzero && class1 == gzero)
	err(talker, "prime equals composite in ifac_sort_one");
      else if (class0 == gdeux)	/* should happen even less */
	scan[2] = (long)class0;	/* use it */
    }
    else			/* shouldn't happen either */
      scan[2] = (long)class0;	/* use it */
  }
  /* else stay with the existing known class0 */
  scan[2] = (long)class1;
  /* in any case, add exponents */
  if (scan[-2] == un && exponent == gun)
    scan[1] = deux;
  else
    scan[1] = laddii((GEN)(scan[-2]), exponent);
  /* move the value over */
  *scan = scan[-3];
  /* null out the vacated slot below */
  *--scan = (long)NULL;
  *--scan = (long)NULL;
  *--scan = (long)NULL;
  /* finally, see whether *where should be pulled in */
  if (scan == *where) *where += 3;
  return 0;
}

/* the following loop around the former doesn't need to check moebius_mode
   because ifac_sort_one() never returns 1 in normal mode */
static long
ifac_resort(GEN *partial, GEN *where)
{
  long lgp = lg(*partial), res = 0;
  GEN scan = *where;

  for (; scan < *partial + lgp; scan += 3)
  {
    if (*scan &&		/* slot occupied */
	!scan[2])		/* with an unknown */
    {
      res |= ifac_sort_one(partial, where, scan);
      if (res) return res;	/* early exit */
    }
  }
  return res;
}

/* sweep downward so we can with luck turn some Qs into Ps */
static void
ifac_whoiswho(GEN *partial, GEN *where, long after_crack)
{
  long lgp = lg(*partial), larger_compos = 0;
  GEN scan, scan_end = *partial + lgp - 3;

  if (DEBUGLEVEL > 4)
  {
    if (!*partial || typ(*partial) != t_VEC)
      err(typeer, "ifac_whoiswho");
    if (lg(*partial) < ifac_initial_length)
      err(talker, "partial impossibly short in ifac_whoiswho");
    if (!(*where) ||
	*where > scan_end ||
        *where < *partial + 3)
      err(talker, "`*where\' out of bounds in ifac_whoiswho");
  }

  if (after_crack == 0) return;
  if (after_crack > 0)
  {
    larger_compos = 1;		/* disable Q-to-P trick */
    scan = *where + 3*(after_crack - 1);
				/* check at most after_crack entries */
    if (scan > scan_end)	/* ooops... */
    {
      err(warner, "avoiding nonexistent factors in ifac_whoiswho");
      scan = scan_end;
    }
  }
  else { larger_compos = 0; scan = scan_end; }

  for (; scan >= *where; scan -= 3)
  {
    if (scan[2])		/* known class of factor */
    {
      if (scan[2] == zero) larger_compos = 1;
      else if (!larger_compos && scan[2] == un)
      {
	if (DEBUGLEVEL >= 3)
	{
	  fprintferr("IFAC: factor %Z\n\tis prime (no larger composite)\n",
		     (long)(**where));
	  fprintferr("IFAC: prime %Z\n\tappears with exponent = %ld\n",
		     (long)(**where), (long)itos((GEN)(*where)[1]));
	}
	scan[2] = deux;
      }	/* no else case */
      continue;
    }
    scan[2] =
      (isprime((GEN)(*scan)) ?
       (larger_compos ? un : deux) : /* un- or finished prime */
       zero);			/* composite */

    if (scan[2] == zero) larger_compos = 1;
    if (DEBUGLEVEL >= 3)
    {
      fprintferr("IFAC: factor %Z\n\tis %s\n",
		 (long)(*scan),
		 (scan[2] == zero ? "composite" : "prime"));
    }
  }
}

/* Here we normally do not check that the first entry is a not-finished
   prime.  Stack management: we may allocate a new exponent */
static long
ifac_divide(GEN *partial, GEN *where)
{
  long lgp = lg(*partial);
  GEN scan = *where + 3;
  long res = 0, exponent, newexp, otherexp;

  if (DEBUGLEVEL > 4)		/* none of these should ever happen */
  {
    if (!*partial || typ(*partial) != t_VEC)
      err(typeer, "ifac_divide");
    if (lg(*partial) < ifac_initial_length)
      err(talker, "partial impossibly short in ifac_divide");
    if (!(*where) ||
	*where > *partial + lgp - 3 ||
        *where < *partial + 3)
      err(talker, "`*where\' out of bounds in ifac_divide");
    if ((*where)[2] != un)
      err(talker, "division by composite or finished prime in ifac_divide");
  }
  if (!(**where))		/* always test just this one */
    err(talker, "division by nothing in ifac_divide");

  newexp = exponent = itos((GEN)((*where)[1]));
  if (exponent > 1 && moebius_mode) return 1;
  /* should've been caught by caller already */

  /* go for it */
  for (; scan < *partial + lgp; scan += 3)
  {
    if (scan[2] != zero) continue; /* the other thing ain't composite */
    otherexp = 0;
    /* let mpdivis divide the other factor in place to keep stack clutter
       minimal */
    while (mpdivis((GEN)(*scan), (GEN)(**where), (GEN)(*scan)))
    {
      if (moebius_mode) return 1; /* immediately */
      if (!otherexp) otherexp = itos((GEN)(scan[1]));
      newexp += otherexp;
    }
    if (newexp > exponent)	/* did anything happen? */
    {
      (*where)[1] = (newexp == 2 ? deux : (long)(stoi(newexp)));
      exponent = newexp;
      if (is_pm1((GEN)(*scan))) /* factor dissolved completely */
      {
	*scan = scan[1] = (long)NULL;
	if (DEBUGLEVEL >= 4)
	  fprintferr("IFAC: a factor was a power of another prime factor\n");
      }
      else if (DEBUGLEVEL >= 4)
      {
	fprintferr("IFAC: a factor was divisible by another prime factor,\n");
	fprintferr("\tleaving a cofactor = %Z\n", (long)(*scan));
      }
      scan[2] = (long)NULL;	/* at any rate it's Unknown now */
      res = 1;
      if (DEBUGLEVEL >= 5)
      {
	fprintferr("IFAC: prime %Z\n\tappears at least to the power %ld\n",
		   (long)(**where), newexp);
      }
    }
  } /* for */
  (*where)[2] = deux;		/* make it a finished prime */
  if (DEBUGLEVEL >= 3)
  {
    fprintferr("IFAC: prime %Z\n\tappears with exponent = %ld\n",
	       (long)(**where), newexp);
  }
  return res;
}


GEN mpqs(GEN N);		/* in src/modules/mpqs.c, maybe a dummy,
				   returns a factor, or a vector of factors,
				   or NULL */

/* The following takes the place of 2.0.9.alpha's find_factor(). */

/* The meaning of the hint changes against 2.0.9.alpha to:
   hint == 0 : Use our own strategy, and this should be the default
   hint & 1  : Avoid mpqs(), use ellfacteur() after pollardbrent()
   hint & 2  : Avoid first-stage ellfacteur() in favour of mpqs()
   (which may still fall back to ellfacteur() if mpqs() is not installed
   or gives up)
   hint & 4  : Avoid even the pollardbrent() stage
   hint & 8  : For size-limited factorization -- do not run any expensive
   stages  (mpqs() and final ellfacteur()); instead, just look for powers. */

/* stack housekeeping:  this routine may create one or more objects  (a new
   factor, or possibly several, and perhaps one or more new exponents > 2) */
static long
ifac_crack(GEN *partial, GEN *where)
{
  long hint, cmp_res, exp1 = 1, exp2 = 1, av;
  GEN factor = NULL, exponent;

  if (DEBUGLEVEL > 4)		/* none of these should ever happen */
  {
    long lgp;
    if (!*partial || typ(*partial) != t_VEC)
      err(typeer, "ifac_crack");
    if ((lgp = lg(*partial)) < ifac_initial_length)
      err(talker, "partial impossibly short in ifac_crack");
    if (!(*where) ||
	*where < *partial + 6 || /* sic -- caller must realloc first */
	*where > *partial + lgp - 3)
      err(talker, "`*where\' out of bounds in ifac_crack");
    if (!(**where) || typ((GEN)(**where)) != t_INT)
      err(typeer, "ifac_crack");
    if ((*where)[2] != zero)
      err(talker, "operand not known composite in ifac_crack");
  }
  hint = itos((GEN)((*partial)[2])) & 15;
  exponent = (GEN)((*where)[1]);

  if (DEBUGLEVEL >= 3)
    fprintferr("IFAC: cracking composite\n\t%Z\n",
	       (long)(**where));

  /* crack squares.  Quite fast due to the initial square residue test */
  if (DEBUGLEVEL >= 4)
    fprintferr("IFAC: checking for pure square\n");
  av = avma;
  while (carrecomplet((GEN)(**where), &factor))
  {
    if (DEBUGLEVEL >= 4)
      fprintferr("IFAC: found %Z =\n\t%Z ^2\n",
		 (long)(**where), (long)factor);
    affii(factor, (GEN)(**where)); avma = av; factor = NULL;
    if (exponent == gun)
      (*where)[1] = deux;
    else if (exponent == gdeux)
    { (*where)[1] = (long)stoi(4); av = avma; }
    else
    { affii(shifti(exponent, 1), (GEN)((*where)[1])); avma = av; }
    exponent = (GEN)((*where)[1]);
    if (moebius_mode) return 0;	/* no need to carry on... */
    exp1 = 2;
  } /* while carrecomplet */

  /* check whether our composite hasn't become prime */
  if (exp1 > 1 && isprime((GEN)(**where)))
  {
    (*where)[2] = un;
    if (DEBUGLEVEL >= 4)
    {
      fprintferr("IFAC: factor %Z\n\tis prime\n",
		 (long)(**where));
      flusherr();
    }
    return 0;			/* bypass subsequent ifac_whoiswho() call */
  }
  /* still composite -- carry on */

  /* MPQS cannot factor prime powers;  check for cubes/5th/7th powers.
     Do this even if MPQS is blocked by hint -- it still serves a useful
     purpose in bounded factorization */
  {
    long mask = 7;
    if (DEBUGLEVEL == 4)
      fprintferr("IFAC: checking for odd power\n");
    /* (At debug levels > 4, is_odd_power() itself prints something more
       informative) */
    av = avma;
    while ((exp1 =		/* assignment */
	    is_odd_power((GEN)(**where), &factor, &mask)))
    {
      if (exp2 == 1) exp2 = exp1; /* remember this after the loop */
      if (DEBUGLEVEL >= 4)
	fprintferr("IFAC: found %Z =\n\t%Z ^%ld\n",
		   (long)(**where), (long)factor, exp1);
      affii(factor, (GEN)(**where)); avma = av; factor = NULL;
      if (exponent == gun)
      { (*where)[1] = (long)stoi(exp1); av = avma; }
      else if (exponent == gdeux)
      { (*where)[1] = (long)stoi(exp1<<1); av = avma; }
      else
      { affii(mulsi(exp1, exponent), (GEN)((*where)[1])); avma = av; }
      exponent = (GEN)((*where)[1]);
      if (moebius_mode) return 0; /* no need to carry on... */
    } /* while is_odd_power */

    if (exp2 > 1)
    {				/* Something nice has happened */
      /* check whether our composite hasn't become prime */
      if (isprime((GEN)(**where)))
      {
        (*where)[2] = un;
	if (DEBUGLEVEL >= 4)
	{
	  fprintferr("IFAC: factor %Z\n\tis prime\n",
		     (long)(**where));
	  flusherr();
	}
	return 0;		/* bypass subsequent ifac_whoiswho() call */
      }
      /* base of power is still composite  (an exceedingly rare case),
	 fall through */
    }
  } /* odd power stage */

  /* pollardbrent() Rho usually gets a first chance */
  if (!(hint & 4))
  {
    if (DEBUGLEVEL >= 4)
      fprintferr("IFAC: trying Pollard-Brent rho method first\n");
    factor = pollardbrent((GEN)(**where));
  } /* Rho stage */

  /* if this didn't work, try one of our high-power beasties */
  if (!factor && !(hint & 2))
  {
    if (DEBUGLEVEL >= 4)
      fprintferr("IFAC: trying Lenstra-Montgomery ECM\n");
    factor = ellfacteur((GEN)(**where), 0); /* do not insist */
  } /* First ECM stage */

  if (!factor && !(hint & 9))
  {
    if (DEBUGLEVEL >= 4)
      fprintferr("IFAC: trying Multi-Polynomial Quadratic Sieve\n");
    factor = mpqs((GEN)(**where));
  } /* MPQS stage */

  if (!factor)
  {
    if (!(hint & 8))		/* still no luck?  force it */
    {
      if (DEBUGLEVEL >= 4)
	fprintferr("IFAC: forcing ECM, may take some time\n");
      factor = ellfacteur((GEN)(**where), 1);
    } /* final ECM stage, guaranteed to succeed */
    else			/* limited factorization */
    {
      if (DEBUGLEVEL >= 2)
      {
	err(warner, "IFAC: unfactored composite declared prime");
	/* don't print it out at level 3 or above, where it would appear
	   several times before and after this message already */
	if (DEBUGLEVEL == 2)
	{
	  fprintferr("\t%Z\n", (long)(**where));
	  flusherr();
	}
      }
      (*where)[2] = un;		/* might as well trial-divide by it... */
      return 1;
    }
  } /* Final ECM stage */

  if (DEBUGLEVEL > 0)
  {
    if (!factor)		/* never reached */
      err(talker, "all available factoring methods failed in ifac_crack");
  }
  if (typ(factor) == t_VEC)	/* delegate this case */
    return ifac_insert_multiplet(partial, where, factor);

  else if (typ(factor) != t_INT)
  {
    fprintferr("IFAC: factorizer returned strange object to ifac_crack\n");
    outerr(factor);
    err(bugparier, "factoring");
  }

  /* got single integer back:  work out the cofactor (in place) */
  if (!mpdivis((GEN)(**where), factor, (GEN)(**where)))
  {
    fprintferr("IFAC: factoring %Z\n", (long)(**where));
    fprintferr("\tyielded `factor\' %Z\n\twhich isn't!\n",
	       (long)factor);
    err(bugparier, "factoring");
  }

  /* the factoring engines report the factor found when DEBUGLEVEL is
     large enough;  let's tell about the cofactor */
  if (DEBUGLEVEL >= 4)
    fprintferr("IFAC: cofactor = %Z\n", (long)(**where));

  /* ok, now `factor' is one factor and **where is the other, find out
     which is larger */
  cmp_res = cmpii(factor, (GEN)(**where));
  if (cmp_res < 0)		/* common case */
  {
    (*where)[2] = (long)NULL;	/* mark cofactor `unknown' */
    (*where)[-1] = (long)NULL;	/* mark factor `unknown' */
    (*where)[-2] =
      isonstack(exponent) ? licopy(exponent) : (long)exponent;
    *where -= 3;
    **where = (long)factor;
    return 2;
  }
  else if (cmp_res == 0)	/* hep, split a square in the middle */
  {
    err(warner,
	"square not found by carrecomplet, ifac_crack recovering");
    cgiv(factor);
    (*where)[2] = (long)NULL;	/* mark the sqrt `unknown' */
    if (exponent == gun)	/* double the exponent */
      (*where)[1] = deux;
    else if (exponent == gdeux)
      (*where)[1] = (long)stoi(4); /* make a new one */
    else			/* overwrite old exponent */
    {
      av = avma;
      affii(shifti(exponent, 1), (GEN)((*where)[1]));
      avma = av;
      /* leave *where unchanged */
    }
    if (moebius_mode) return 0;
    return 1;
  }
  else				/* factor > cofactor, rearrange */
  {
    (*where)[2] = (long)NULL;	/* mark factor `unknown' */
    (*where)[-1] = (long)NULL;	/* mark cofactor `unknown' */
    (*where)[-2] =
      isonstack(exponent) ? licopy(exponent) : (long)exponent;
    *where -= 3;
    **where = (*where)[3];	/* move cofactor pointer to lowest slot */
    (*where)[3] = (long)factor;	/* save factor */
    return 2;
  }
}

/* the following doesn't collect garbage;  caller's caller should do it
   (which means ifac_main()).  No diagnostics either, the factoring engine
   should have printed what it found when DEBUGLEVEL>=4 or so.  Note facvec
   contains slots of three components per factor;  repeated factors are
   expressly allowed  (and their classes shouldn't contradict each other
   whereas their exponents will be added up) */
static long
ifac_insert_multiplet(GEN *partial, GEN *where, GEN facvec)
{
  long j,k=1, lfv=lg(facvec)-1, nf=lfv/3, room=(long)(*where-*partial);
  /* one of the factors will go into the *where slot, so room is now
     3 times the number of slots we can use */
  long needroom = lfv - room;
  GEN sorted, auxvec = cgetg(nf+1, t_VEC), factor;
  long exponent = itos((GEN)((*where)[1])); /* the old exponent */
  GEN newexp;

  if (DEBUGLEVEL >= 5)
    fprintferr("IFAC: incorporating set of %ld factors%s\n",
	       nf, (DEBUGLEVEL >=6 ? "..." : ""));
  if (needroom > 0)
    ifac_realloc(partial, where, lg(*partial) + needroom + 3);
  /* one extra slot for paranoia, errm, future use */

  /* create sort permutation from the values of the factors */
  for (j=nf; j; j--) auxvec[j] = facvec[3*j-2]; /* just the pointers */
  sorted = sindexsort(auxvec);
  /* and readjust the result for the triple spacing */
  for (j=nf; j; j--) sorted[j] = 3*sorted[j]-2;
  if (DEBUGLEVEL >= 6)
    fprintferr("\tsorted them...\n");

  /* store factors, beginning at *where, and catching any duplicates */
  **where = facvec[sorted[nf]];
  if ((newexp = (GEN)(facvec[sorted[nf]+1])) != gun) /* new exponent > 1 */
  {
    if (exponent == 1)
      (*where)[1] = isonstack(newexp) ? licopy(newexp) : (long)newexp;
    else
      (*where)[1] = lmulsi(exponent, newexp);
  } /* if new exponent is 1, the old exponent already in place will do */
  (*where)[2] = facvec[sorted[nf]+2]; /* copy class */
  if (DEBUGLEVEL >= 6)
    fprintferr("\tstored (largest) factor no. %ld...\n", nf);

  for (j=nf-1; j; j--)
  {
    factor = (GEN)(facvec[sorted[j]]);
    if (egalii(factor, (GEN)(**where)))
    {
      if (DEBUGLEVEL >= 6)
	fprintferr("\tfactor no. %ld is a duplicate%s\n",
		   j, (j>1 ? "..." : ""));
      /* update exponent, ignore class which would already have been set,
	 and then forget current factor */
      if ((newexp = (GEN)(facvec[sorted[j]+1])) != gun) /* new exp > 1 */
      {				/* now we have at least 3 */
	(*where)[1] = laddii((GEN)((*where)[1]),
			     mulsi(exponent, newexp));
      }
      else
      {
	if ((*where)[1] == un && exponent == 1)
	  (*where)[1] = deux;
	else
	  (*where)[1] = laddsi(exponent, (GEN)((*where)[1]));
	/* not safe to add 1 in place -- that might overwrite gdeux,
	   with `interesting' consequences */
      }
      if (moebius_mode) return 0; /* stop now, but with exponent updated */
      continue;
    }
    (*where)[-1] = facvec[sorted[j]+2];	/* class as given */
    if ((newexp = (GEN)(facvec[sorted[j]+1])) != gun) /* new exp > 1 */
    {
      if (exponent == 1 && newexp == gdeux)
	(*where)[-2] = deux;
      else			/* exponent*newexp > 2 */
	(*where)[-2] = lmulsi(exponent, newexp);
    }
    else
    {
      (*where)[-2] = (exponent == 1 ? un :
		      (exponent == 2 ? deux :
		       (long)stoi(exponent))); /* inherit parent's exponent */
    }
    (*where)[-3] = isonstack(factor) ? licopy(factor) : (long)factor;
				/* keep components younger than *partial */
    *where -= 3;
    k++;
    if (DEBUGLEVEL >= 6)
      fprintferr("\tfactor no. %ld was unique%s\n",
		 j, (j>1 ? " (so far)..." : ""));
  }
  /* make the `sorted' object safe for garbage collection  (probably not a
     problem, since it should be in the garbage zone from everybody's
     perspective, but it's easy to do it) */
  *sorted = evaltyp(t_INT) | evallg(nf+1);
  return k;
}

static GEN
ifac_main(GEN *partial)
{
  /* leave the basic error checking to ifac_find() */
  GEN here = ifac_find(partial, partial);
  long res, nf;

  /* if nothing left, return gun */
  if (!here) return gun;

  /* if we are in Moebius mode and have already detected a repeated factor,
     stop right here.  Shouldn't really happen */
  if (moebius_mode && here[1] != un)
  {
    if (DEBUGLEVEL >= 3)
    {
      fprintferr("IFAC: main loop: repeated old factor\n\t%Z\n",
		 (long)(*here));
      flusherr();
    }
    return gzero;
  }

  /* loop until first entry is a finished prime.  May involve reallocations
     and thus updates of *partial */
  while (here[2] != deux)
  {
    /* if it's unknown, something has gone wrong;  try to recover */
    if (!(here[2]))
    {
      err(warner, "IFAC: unknown factor seen in main loop");
      res = ifac_resort(partial, &here);
      if (res) return gzero;	/* can only happen in Moebius mode */
      ifac_whoiswho(partial, &here, -1);
      /* defrag for good measure */
      ifac_defrag(partial, &here);
      continue;
    }
    /* if it's composite, crack it */
    if (here[2] == zero)
    {
      /* make sure there's room for another factor */
      if (here < *partial + 6)
      {				/* try defrag first */
	ifac_defrag(partial, &here);
	if (here < *partial + 6) /* no luck */
	{
	  ifac_realloc(partial, &here, 1); /* guaranteed to work */
	  /* Unfortunately, we can't do a garbage collection here since we
	     know too little about where in the stack the old components
	     were. */
	}
      }
      nf = ifac_crack(partial, &here);
      if (moebius_mode && here[1] != un) /* that was a power */
      {
	if (DEBUGLEVEL >= 3)
	{
	  fprintferr("IFAC: main loop: repeated new factor\n\t%Z\n",
		     (long)(*here));
	  flusherr();
	}
	return gzero;
      }
      /* deal with the new unknowns.  No resort, since ifac_crack will
	 already have sorted them */
      ifac_whoiswho(partial, &here, nf);
      continue;
    }
    /* if it's prime but not yet finished, finish it */
    if (here[2] == un)
    {
      res = ifac_divide(partial, &here);
      if (res)
      {
	if (moebius_mode)
	{
	  if (DEBUGLEVEL >= 3)
	  {
	    fprintferr("IFAC: main loop: another factor was divisible by\n");
	    fprintferr("\t%Z\n", (long)(*here));
	    flusherr();
	  }
	  return gzero;
	}
	ifac_defrag(partial, &here);
	(void)(ifac_resort(partial, &here)); /* sort new cofactors down */
	/* it doesn't matter right now whether this finds a repeated factor,
	   since we never get to this point in Moebius mode */
	ifac_defrag(partial, &here); /* resort may have created new gaps */
	ifac_whoiswho(partial, &here, -1);
      }
      continue;
    }
    /* there are no other cases, never reached */
    err(talker, "non-existent factor class in ifac_main");
  } /* while */
  if (moebius_mode && here[1] != un)
  {
    if (DEBUGLEVEL >= 3)
    {
      fprintferr("IFAC: after main loop: repeated old factor\n\t%Z\n",
		 (long)(*here));
      flusherr();
    }
    return gzero; /* just a safety net */
  }
  if (DEBUGLEVEL >= 4)
  {
    long nf = (*partial + lg(*partial) - here - 3)/3;
    if (nf)
      fprintferr("IFAC: main loop: %ld factor%s left\n",
		 nf, (nf>1 ? "s" : ""));
    else
      fprintferr("IFAC: main loop: this was the last factor\n");
    flusherr();
  }
  return here;
}

/* Caller of the following should worry about stack management, it makes
   a rather shameless mess :^) */
GEN
ifac_primary_factor(GEN *partial, long *exponent)
{
  GEN here = ifac_main(partial);
  GEN res;

  if (here == gun) { *exponent = 0; return gun; }
  else if (here == gzero) { *exponent = 0; return gzero; }

  res = icopy((GEN)(*here));
  *exponent = itos((GEN)(here[1]));
  here[2] = here[1] = *here = (long)NULL;
  return res;
}

/* encapsulated routines */

/* prime/exponent pairs need to appear contiguously on the stack, but we
   also need to have our data structure somewhere, and we don't know in
   advance how many primes will turn up.  The following discipline achieves
   this:  When ifac_decomp() is called, n should point at an object older
   than the oldest small prime/exponent pair  (auxdecomp0() guarantees
   this easily since it mpdivis()es any divisors it discovers off its own
   copy of the original N).  We allocate sufficient space to accommodate
   a several pairs -- eleven pairs ought to fit in a space not much larger
   than n itself -- before calling ifac_start().  If we manage to complete
   the factorization before we run out of space, we free the data structure
   and cull the excess reserved space before returning.  When we do run out,
   we have to leapfrog to generate more  (guesstimating the requirements
   from what is left in the partial factorization structure);  room for
   fresh pairs is allocated at the head of the stack, followed by an
   ifac_realloc() to reconnect the data structure and move it out of the
   way, followed by a garbage collection to connect the new pairs space
   to the old one.-- This whole affair translates into a surprisingly
   compact little routine. */

#define ifac_overshoot 64	/* lgefint(n)+64 words reserved */

long
ifac_decomp(GEN n, long hint)
{
  long tf=lgefint(n), av=avma, lim=(av+bot)>>1;
  long nb=0;
  GEN part, here, workspc = cgeti(tf + ifac_overshoot), pairs = (GEN)av;
  /* workspc will be doled out by us in pairs of smaller t_INTs */
  long tetpil = avma;		/* remember head of workspc zone */

  if (!n || typ(n) != t_INT) err(typeer, "ifac_decomp");
  if (!signe(n) || tf < 3) err(talker, "factoring 0 in ifac_decomp");

  part = ifac_start(n, 0, hint);
  here = ifac_main(&part);

  while (here != gun)
  {
    long lf=lgefint((GEN)(*here));
    if (pairs - workspc < lf + 3) /* out of room, leapfrog */
    {
      long lt=lf+3, tetpil1=avma;
      GEN *gsav[2], here1 = here+3, end = part+lg(part);
      if(DEBUGMEM>1) err(warnmem,"[1] ifac_decomp");
      /* guesstimate how much more will be needed */
      while (here1 < end)
      {
	lt += lgefint((GEN)(*here1)) + 3; /* allow for exponent */
	if (!here1[2] || here1[2] == zero) /* (unknown or) composite... */
	  lt+=12;
	/* ...allow for 2 further prime/exponent pairs  (reserving for each
	   one extra significand word, 2 codewords, and a 3-word exponent) */
	here1 += 3;
      }
      workspc = cgeti(lt + (ifac_overshoot>>2)); /* + some paranoia room */
      ifac_realloc(&part, &here, 0);
      gsav[0] = &workspc; gsav[1] = &part;
      gerepilemanysp(tetpil, tetpil1, gsav, 2);
      here = ifac_find(&part, &part);
      tetpil = (long)workspc;
    }
    /* room enough now */
    nb++;
    pairs -= lf;
    *pairs = evaltyp(t_INT) | evallg(lf);
    affii((GEN)(*here), pairs);
    pairs -= 3;
    *pairs = evaltyp(t_INT) | evallg(3);
    affii((GEN)(here[1]), pairs);
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      if(DEBUGMEM>1) err(warnmem,"[2] ifac_decomp");
      ifac_realloc(&part, &here, 0);
      part = gerepileupto(tetpil, part);
    }
  }
  avma = (long)pairs;
  if (DEBUGLEVEL >= 3)
  {
    fprintferr("IFAC: found %ld large prime (power) factor%s.\n",
	       nb, (nb>1? "s": ""));
    flusherr();
  }
  return nb;
}

long
ifac_moebius(GEN n, long hint)
{
  long mu=1, av=avma, lim=(av+bot)>>1;
  GEN part = ifac_start(n, 1, hint);
  GEN here = ifac_main(&part);

  while (here != gun && here != gzero)
  {
    if (itos((GEN)(here[1])) > 1)
    { here = gzero; break; }	/* shouldn't happen */
    mu = -mu;
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      if(DEBUGMEM>1) err(warnmem,"ifac_moebius");
      ifac_realloc(&part, &here, 0);
      part = gerepileupto(av, part);
    }
  }
  avma = av;
  return (here == gun ? mu : 0);
}

long
ifac_issquarefree(GEN n, long hint)
{
  long av=avma, lim=(av+bot)>>1;
  GEN part = ifac_start(n, 1, hint);
  GEN here = ifac_main(&part);

  while (here != gun && here != gzero)
  {
    if (itos((GEN)(here[1])) > 1)
    { here = gzero; break; }	/* shouldn't happen */
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      if(DEBUGMEM>1) err(warnmem,"ifac_issquarefree");
      ifac_realloc(&part, &here, 0);
      part = gerepileupto(av, part);
    }
  }
  avma = av;
  return (here == gun ? 1 : 0);
}

long
ifac_omega(GEN n, long hint)
{
  long omega=0, av=avma, lim=(av+bot)>>1;
  GEN part = ifac_start(n, 0, hint);
  GEN here = ifac_main(&part);

  while (here != gun)
  {
    omega++;
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      if(DEBUGMEM>1) err(warnmem,"ifac_omega");
      ifac_realloc(&part, &here, 0);
      part = gerepileupto(av, part);
    }
  }
  avma = av;
  return omega;
}

long
ifac_bigomega(GEN n, long hint)
{
  long Omega=0, av=avma, lim=(av+bot)>>1;
  GEN part = ifac_start(n, 0, hint);
  GEN here = ifac_main(&part);

  while (here != gun)
  {
    Omega += itos((GEN)(here[1]));
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      if(DEBUGMEM>1) err(warnmem,"ifac_bigomega");
      ifac_realloc(&part, &here, 0);
      part = gerepileupto(av, part);
    }
  }
  avma = av;
  return Omega;
}

GEN
ifac_totient(GEN n, long hint)
{
  GEN res = cgeti(lgefint(n));
  long exponent, av=avma, tetpil, lim=(av+bot)>>1;
  GEN phi = gun;
  GEN part = ifac_start(n, 0, hint);
  GEN here = ifac_main(&part);

  while (here != gun)
  {
    phi = mulii(phi, addsi(-1, (GEN)(*here)));
    if (here[1] != un)
      if (here[1] == deux)
      {
	phi = mulii(phi, (GEN)(*here));
      }
      else
      {
	exponent = itos((GEN)(here[1]));
	phi = mulii(phi, gpowgs((GEN)(*here), exponent-1));
      }
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      GEN *gsav[2];
      if(DEBUGMEM>1) err(warnmem,"ifac_totient");
      tetpil = avma;
      ifac_realloc(&part, &here, 0);
      phi = icopy(phi);
      gsav[0] = &phi; gsav[1] = &part;
      gerepilemanysp(av, tetpil, gsav, 2);
      /* don't try to preserve here, safer to pick it up again
	 (and ifac_find does a lot of sanity checking at high
	 debuglevels) */
      here = ifac_find(&part, &part);
    }
  }
  affii(phi, res);
  avma = av;
  return res;
}

GEN
ifac_numdiv(GEN n, long hint)
{
  /* we don't preallocate since it's too hard to guess the right
     size here */
  GEN res;
  long av=avma, tetpil, lim=(av+bot)>>1;
  GEN exponent, tau = gun;
  GEN part = ifac_start(n, 0, hint);
  GEN here = ifac_main(&part);

  while (here != gun)
  {
    exponent = (GEN)(here[1]);
    tau = mulii(tau, addsi(1, exponent));
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      GEN *gsav[2];
      if(DEBUGMEM>1) err(warnmem,"ifac_numdiv");
      tetpil = avma;
      ifac_realloc(&part, &here, 0);
      tau = icopy(tau);
      gsav[0] = &tau; gsav[1] = &part;
      gerepilemanysp(av, tetpil, gsav, 2);
      /* (see ifac_totient()) */
      here = ifac_find(&part, &part);
    }
  }
  tetpil = avma;
  res = icopy(tau);
  return gerepile(av, tetpil, res);
}

GEN
ifac_sumdiv(GEN n, long hint)
{
  /* don't preallocate */
  GEN res;
  long exponent, av=avma, tetpil, lim=(av+bot)>>1;
  GEN contrib, sigma = gun;
  GEN part = ifac_start(n, 0, hint);
  GEN here = ifac_main(&part);

  while (here != gun)
  {
    exponent = itos((GEN)(here[1]));
    contrib = addsi(1, (GEN)(*here));
    for (; exponent > 1; exponent--)
      contrib = addsi(1, mulii((GEN)(*here), contrib));
    sigma = mulii(sigma, contrib);
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      GEN *gsav[2];
      if(DEBUGMEM>1) err(warnmem,"ifac_sumdiv");
      tetpil = avma;
      ifac_realloc(&part, &here, 0);
      sigma = icopy(sigma);
      gsav[0] = &sigma; gsav[1] = &part;
      gerepilemanysp(av, tetpil, gsav, 2);
      /* (see ifac_totient()) */
      here = ifac_find(&part, &part);
    }
  }
  tetpil = avma;
  res = icopy(sigma);
  return gerepile(av, tetpil, res);
}

/* k should be positive, and indeed it had better be > 1  (not checked).
   The calling function knows what to do with the other cases. */
GEN
ifac_sumdivk(GEN n, long k, long hint)
{
  /* don't preallocate */
  GEN res;
  long exponent, av=avma, tetpil, lim=(av+bot)>>1;
  GEN contrib, q, sigma = gun;
  GEN part = ifac_start(n, 0, hint);
  GEN here = ifac_main(&part);

  while (here != gun)
  {
    exponent = itos((GEN)(here[1]));
    q = gpowgs((GEN)(*here), k);
    contrib = addsi(1, q);
    for (; exponent > 1; exponent--)
      contrib = addsi(1, mulii(q, contrib));
    sigma = mulii(sigma, contrib);
    here[2] = here[1] = *here = (long)NULL;
    here = ifac_main(&part);
    if (low_stack(lim, (av+bot)>>1))
    {
      GEN *gsav[2];
      if(DEBUGMEM>1) err(warnmem,"ifac_sumdivk");
      tetpil = avma;
      ifac_realloc(&part, &here, 0);
      sigma = icopy(sigma);
      gsav[0] = &sigma; gsav[1] = &part;
      gerepilemanysp(av, tetpil, gsav, 2);
      /* (see ifac_totient()) */
      here = ifac_find(&part, &part);
    }
  }
  tetpil = avma;
  res = icopy(sigma);
  return gerepile(av, tetpil, res);
}
